/*
**             Copyright 2017 by Kvaser AB, Molndal, Sweden
**                         http://www.kvaser.com
**
** This software is dual licensed under the following two licenses:
** BSD-new and GPLv2. You may use either one. See the included
** COPYING file for details.
**
** License: BSD-new
** ==============================================================================
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the <organization> nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
** BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
**
**
** License: GPLv2
** ==============================================================================
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
**
**
** IMPORTANT NOTICE:
** ==============================================================================
** This source code is made available for free, as an open license, by Kvaser AB,
** for use with its applications. Kvaser AB does not accept any liability
** whatsoever for any third party patent or other immaterial property rights
** violations that may result from any usage of this source code, regardless of
** the combination of source code and various applications that it can be used
** in, or with.
**
** -----------------------------------------------------------------------------
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "kvaDbLib.h"
#include "CANdb.h"
#include "CANdbDbc.h"
#include "debug.h"


static int roundi(double x) { return (int)(x+(x>0.0?0.5:-0.5)); }
static int roundu(double x) { return (unsigned int)(x+0.5); }
static int kvastricmp(const char *s1, const char *s2)
{
  int c1, c2;
  for(;;) {
    c1 = tolower((unsigned char) *s1++);
    c2 = tolower((unsigned char) *s2++);
    if (c1 == 0 || c1 != c2) {
      return c1 - c2;
    }
  }
}
#ifdef DEBUG
static void dbg_printf(const char *s, ...);
#endif
//===========================================================================
KvaDbStatus WINAPI kvaDbOpen(KvaDbHnd *dh)
{
  if (!dh) return kvaDbErr_Param;
  CANdbCluster * dc = new CANdbCluster ();
  *dh = (KvaDbHnd) dc;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbAddFile(KvaDbHnd dh, const char *filename)
{
  if (!filename) return kvaDbErr_Param;
  if (!dh) return kvaDbErr_Param;

  CANdbCluster * dc = (CANdbCluster *)dh;
  CANdb *db = new CANdb;

  CANdbDBC dbcio (filename);
  if (dbcio.read_file (db) != 0) {
    delete db;
    return kvaDbErr_DbFileOpen;
  }

  db->set_name (dbcio.get_db_name ());
  db->set_filename (filename);
  dc->add_db (db);

  return kvaDbOK;
}

//===========================================================================
// Used to define a "file name" when we don't have any file that the
// data have been read from.
KvaDbStatus WINAPI kvaDbSetDummyFileName(KvaDbHnd dh, const char *filename)
{
  if (!filename) return kvaDbErr_Param;
  if (!dh) return kvaDbErr_Param;

  CANdbCluster * dc = (CANdbCluster *)dh;
  CANdb *db = new CANdb;

  db->set_filename (filename);
  db->set_name (filename);
  dc->add_db (db);

  return kvaDbOK;
}

//===========================================================================
// Create a database. Can both read from a file and set the database name.
KvaDbStatus WINAPI kvaDbCreate (KvaDbHnd dh, const char *localName, const char *filename)
{
  if (!dh) return kvaDbErr_Param;
  if (!localName && !filename) return kvaDbErr_Param;

  if (localName && !filename) {
    return kvaDbSetDummyFileName(dh, localName);
  }
  else if (!localName && filename) {
    return kvaDbAddFile(dh, filename);
  }
  else {
    CANdbCluster * dc = (CANdbCluster *)dh;
    CANdb *db = new CANdb;

    CANdbDBC dbcio (filename);
    if (dbcio.read_file (db) != 0) {
      delete db;
      return kvaDbErr_DbFileOpen;
    }
    db->set_name (localName);
    db->set_filename (filename);
    dc->add_db (db);
    return kvaDbOK;
  }
}

//===========================================================================
KvaDbStatus WINAPI kvaDbClose(KvaDbHnd dh)
{
  if (!dh) return kvaDbErr_Param;
  CANdbCluster * dc = (CANdbCluster *)dh;
  if (dc) delete dc;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbWriteFile(KvaDbHnd dh, char* filename)
{
  if (!filename) return kvaDbErr_Param;
  if (!dh) return kvaDbErr_Param;

  CANdbCluster * dc = (CANdbCluster *)dh;
  if (!dc) return kvaDbErr_NoDatabase;
  if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *candb = dc->get_first_candb ();
  if (!candb) return kvaDbErr_NoDatabase;

  CANdbDBC dbcio (filename);
  if (dbcio.save_file (candb) != 0) {
    return kvaDbErr_DbFileOpen;
  }

  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbReadFile(KvaDbHnd dh, char* filename)
{
  if (!filename) return kvaDbErr_Param;
  if (!dh) return kvaDbErr_Param;

  CANdbCluster * dc = (CANdbCluster *)dh;
  CANdb *db = new CANdb;

  CANdbDBC dbcio (filename);
  if (dbcio.read_file (db) != 0) {
    delete db;
    return kvaDbErr_DbFileOpen;
  }

  db->set_name (dbcio.get_db_name ());
  db->set_filename (filename);
  dc->add_db (db);

  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetFlags(KvaDbHnd dh, unsigned int *flags)
{
  if (!dh) return kvaDbErr_Param;
  if (!flags) return kvaDbErr_Param;
  CANdbCluster * dc = (CANdbCluster *)dh;
  *flags = 0;
  if (!dc) return kvaDbErr_NoDatabase;
  if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *candb = dc->get_first_candb ();
  if (!candb) return kvaDbErr_NoDatabase;
  if (candb->is_j1939()) *flags |= KVADB_DATABASE_J1939;
  return kvaDbOK;
}

#define KVADB_NUM_PROTOCOLS (kvaDb_ProtocolUnknown + 1)
static const char *KVADBPROTOCOL_STRINGS[] = {"CAN", "Van", "Lin", "MOST", "FlexRay", "BEAN",
                                                 "Ethernet", "AFDX", "J1708", "CAN FD", ""};
CompilerAssert(sizeof(KVADBPROTOCOL_STRINGS) == KVADB_NUM_PROTOCOLS * sizeof(char*));

static const char *KVADBPROTOCOL_ATTRNAME = "BusType";
static const KvaDbProtocolProperties KVADB_PROTOCOL_PROPERTIES[KVADB_NUM_PROTOCOLS] = \
               {{16, 32}, {0, 0}, {8, 32}, {0, 0}, {0, 0},  \
               {0, 0}, {0, 0}, {0, 0}, {0, 0}, {16, 64}, {0, 0}};

//===========================================================================
KvaDbStatus WINAPI kvaDbGetProtocolProperties(KvaDbProtocolType prot, KvaDbProtocolProperties *prop)
{
  if (NULL == prop || (int)prot >= KVADB_NUM_PROTOCOLS) return kvaDbErr_Param;
  *prop = KVADB_PROTOCOL_PROPERTIES[prot];
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbMsgDlcToBytes(KvaDbProtocolType prot, unsigned int dlc, unsigned int *numBytes) {
  if (NULL == numBytes || (int)prot >= KVADB_NUM_PROTOCOLS)
    return kvaDbErr_Param;
  if (0 != KVADB_PROTOCOL_PROPERTIES[prot].maxMessageDlc && dlc > (KVADB_PROTOCOL_PROPERTIES[prot].maxMessageDlc - 1))
    return kvaDbErr_Param;
  switch (prot) {
    case kvaDb_ProtocolCan:
      *numBytes = (dlc > 8? 8: dlc);
      break;
      //intentional fall-through (dlc <= 8)
    case kvaDb_ProtocolCanFD:
      if (dlc > 8) {
        switch (dlc) {
          case 0x9: *numBytes = 12; break;
          case 0xa: *numBytes = 16; break;
          case 0xb: *numBytes = 20; break;
          case 0xc: *numBytes = 24; break;
          case 0xd: *numBytes = 32; break;
          case 0xe: *numBytes = 48; break;
          case 0xf: *numBytes = 64; break;
        }
      }
      else {
        *numBytes = dlc;
      }
      break;
    default:
      return kvaDbErr_Param;
  }
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbBytesToMsgDlc(KvaDbProtocolType prot, unsigned int numBytes, unsigned int *dlc) {
  if (0 == numBytes || (int)prot >= KVADB_NUM_PROTOCOLS)
    return kvaDbErr_Param;
  switch (prot) {
    case kvaDb_ProtocolCan:
      if (numBytes > 8)
        return kvaDbErr_Param;
      *dlc = numBytes;
      break;
    case kvaDb_ProtocolCanFD:
      if (numBytes <= 8)
        *dlc = numBytes;
      else
        switch(numBytes) {
          case 12: *dlc = 0x9; break;
          case 16: *dlc = 0xa; break;
          case 20: *dlc = 0xb; break;
          case 24: *dlc = 0xc; break;
          case 32: *dlc = 0xd; break;
          case 48: *dlc = 0xe; break;
          case 64: *dlc = 0xf; break;
          default: return kvaDbErr_Param;
        }
      break;
    default:
      return kvaDbErr_Param;
  }
  if (0 != KVADB_PROTOCOL_PROPERTIES[prot].maxMessageDlc && *dlc > (KVADB_PROTOCOL_PROPERTIES[prot].maxMessageDlc - 1))
    return kvaDbErr_Param;
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetProtocol(KvaDbHnd dh, KvaDbProtocolType *prot)
{
  if (NULL == dh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;
  if (dc->get_candb_count() > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb* dbc = dc->get_first_candb();
  if (NULL == dbc) return kvaDbErr_NoDatabase;
  KvaDbProtocolType ret_val = kvaDb_ProtocolUnknown;

  CANdbAttributeList *attr_list = dbc->get_attributes();
  if (NULL != attr_list) {
    const CANdbAttribute *prot_attr = attr_list->find_by_name(KVADBPROTOCOL_ATTRNAME);
    if (NULL != prot_attr && CANDB_ATTR_TYPE_STRING == prot_attr->get_type()) {
      const char *prot_val = prot_attr->get_string_value();
      for (int i = 0; i < KVADB_NUM_PROTOCOLS; i++) {
        if (0 == kvastricmp(prot_val, KVADBPROTOCOL_STRINGS[i])) {
          ret_val = static_cast<KvaDbProtocolType>(i);
          break;
        }
    }
  }
  }

  *prot = ret_val;
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetProtocol(KvaDbHnd dh, KvaDbProtocolType prot)
{
  if (NULL == dh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;
  if (dc->get_candb_count() > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb* dbc = dc->get_first_candb();
  if (NULL == dbc) return kvaDbErr_NoDatabase;
  CANdbAttributeList *attr_list = dbc->get_attributes();
  if (NULL == attr_list) return kvaDbErr_NoAttrib;
  CANdbAttribute *prot_attr = attr_list->find_by_name(KVADBPROTOCOL_ATTRNAME);
  if (NULL == prot_attr && kvaDb_ProtocolUnknown != prot)  {
    CANdbAttributeDefinition *bustype_def = dbc->find_attribute_definition_by_name(KVADBPROTOCOL_ATTRNAME);
    if (NULL == bustype_def) {
      bustype_def = new CANdbAttributeDefinition();
      bustype_def->set_owner(CANDB_ATTR_OWNER_DB);
      bustype_def->set_name(KVADBPROTOCOL_ATTRNAME);
      bustype_def->set_type(CANDB_ATTR_TYPE_STRING);
      dbc->insert_attribute_definition(bustype_def);
    }
    CANdbAttribute *new_attr = new CANdbAttribute(bustype_def);
    attr_list->insert(new_attr);
    prot_attr = new_attr;
  }
  else if (CANDB_ATTR_TYPE_STRING != prot_attr->get_type()) {
    return kvaDbErr_DatabaseInternal;
    //Attribute is (un)intentionally corrupted. Or it is some future incompatible version of dbc.
  }
  if (NULL != prot_attr) {
    prot_attr->set_string_value(KVADBPROTOCOL_STRINGS[prot]); //note earlier CompilerAssert
  }

  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetFirstMsg(KvaDbHnd dh, KvaDbMessageHnd *mh)
{
  if (!dh || !mh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;
  if (!dc) return kvaDbErr_NoDatabase;
  CANdbMessage *m = dc->get_first_message();
  if (!m) return kvaDbErr_NoMsg;
  *mh = (KvaDbMessageHnd)m;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetNextMsg(KvaDbHnd dh, KvaDbMessageHnd *mh)
{
  if (!dh || !mh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;

  *mh = (KvaDbMessageHnd)(dc->get_next_message());
  if (!*mh) return kvaDbErr_NoMsg;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgById(KvaDbHnd dh, unsigned int id, KvaDbMessageHnd *mh)
{
  if (!dh || !mh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;

  CANdbMessage *m = dc->find_message_by_id(id);
  if (!m) return kvaDbErr_NoMsg;
  *mh = (KvaDbMessageHnd)m;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgByName(KvaDbHnd dh, const char *msg_name, KvaDbMessageHnd *mh)
{
  if (!dh || !mh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;

  CANdbMessage *m = dc->find_message_by_name(msg_name);
  if (!m) return kvaDbErr_NoMsg;
  *mh = (KvaDbMessageHnd)m;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgName(KvaDbMessageHnd mh, char *buf, size_t buflen)
{
  if (!mh || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  if (m && m->get_name()) strncpy(buf, m->get_name(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgQualifiedName(KvaDbMessageHnd mh, char *buf, size_t buflen)
{
  if (!mh || !buf) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  m->get_qualified_name(buf, (int) buflen);
  buf[buflen-1] = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgComment(KvaDbMessageHnd mh, char *buf, size_t buflen)
{
  if (!mh || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  if (m && m->get_comment()) strncpy(buf, m->get_comment(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgId(KvaDbMessageHnd mh, unsigned int *id, unsigned int *flags)
{
  if (!mh || !id || !flags) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  // qqq ska versta bitten vara satt? eller ska det vara i flags?
  *id = m->get_id_and_ext();
  *flags = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgDlc(KvaDbMessageHnd mh, int *dlc)
{
  if (!mh || !dlc) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  *dlc = m->get_dlc();
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetMsgSendNode(KvaDbMessageHnd mh, KvaDbNodeHnd *nh)
{
  if (!mh || !nh) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;

  CANdbNode *n = m->get_send_node();

  if (!n) return kvaDbErr_NoNode;
  *nh = (KvaDbNodeHnd)n;
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbAddMsg(KvaDbHnd dh, KvaDbMessageHnd *mh)
{
  CANdbCluster *dc = (CANdbCluster *)dh;
  KvaDbStatus stat = kvaDbOK;

  if (!dh || !mh ) return kvaDbErr_Param;

  //if (!dc) return kvaDbErr_NoDatabase;
  //if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *candb = dc->get_first_candb ();
  if (!candb) return kvaDbErr_NoDatabase;

  CANdbMessage *m = new CANdbMessage();
  if (candb->insert_message(m) < 0) {
    stat = kvaDbErr_Param;
    delete m;
    *mh = 0;
  } else {
    m->set_name("");  // qqq or autogenerate?
    m->set_id(0);     // qqq or autogenerate?
    *mh = (KvaDbMessageHnd)m;
  }
  return stat;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbDeleteMsg(KvaDbHnd dh, KvaDbMessageHnd mh)
{
  CANdbCluster *dc = (CANdbCluster *)dh;
  KvaDbStatus stat = kvaDbOK;
  CANdbMessage *m = (CANdbMessage*)mh;

  if (!dh || !mh) return kvaDbErr_Param;

  //if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *candb = dc->get_first_candb ();
  if (!candb) return kvaDbErr_NoDatabase;
  bool success = false;

  while (candb) {
    if (candb->delete_message(m) == 0) {
      //stat = kvaDbErr_DatabaseInternal;
      success = true;
      break;
    }
    candb = dc->get_next_candb ();
  }
  if (!success) stat = kvaDbErr_DatabaseInternal;
  return stat;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetFlags(KvaDbHnd /*dh*/, unsigned int /*flags*/)
{
  // qqq what are these flags?
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetMsgName(KvaDbMessageHnd mh, char *buf)
{
  KvaDbStatus stat = kvaDbOK;
  CANdbMessage *m = (CANdbMessage*)mh;

  if (!mh || !buf) return kvaDbErr_Param;

  m->set_name(buf);
  return stat;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetMsgComment(KvaDbMessageHnd mh, char *buf)
{
  KvaDbStatus stat = kvaDbOK;
  CANdbMessage *m = (CANdbMessage*)mh;

  if (!mh || !buf) return kvaDbErr_Param;

  m->set_comment(buf);
  return stat;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetMsgId(KvaDbMessageHnd mh, unsigned int id, unsigned int flags)
{
  KvaDbStatus stat = kvaDbOK;
  CANdbMessage *m = (CANdbMessage*)mh;

  if (!mh) return kvaDbErr_Param;

  m->set_id(id & ~CANDB_EXT_FLAG);
  m->set_extended(id & CANDB_EXT_FLAG? true: false);
  (void)flags; // qqq
  return stat;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetMsgDlc(KvaDbMessageHnd mh, int dlc)
{
  KvaDbStatus stat = kvaDbOK;
  CANdbMessage *m = (CANdbMessage*)mh;

  if (!mh) return kvaDbErr_Param;

  m->set_dlc(dlc);
  return stat;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetMsgSendNode(KvaDbMessageHnd mh, KvaDbNodeHnd nh)
{
  if (!mh || !nh) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  CANdbNode *n = (CANdbNode*)nh;
  m->set_send_node(n);
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetFirstSignal(KvaDbMessageHnd mh, KvaDbSignalHnd *sh)
{
  if (!mh || !sh) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  CANdbSignal *s = m->get_first_signal();
  *sh = (KvaDbSignalHnd)s;
  if (!s) return kvaDbErr_NoMsg;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetNextSignal(KvaDbMessageHnd mh, KvaDbSignalHnd *sh)
{
  if (!mh || !sh) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;
  CANdbSignal *s = m->get_next_signal();
  *sh = (KvaDbSignalHnd)s;
  if (!s) return kvaDbErr_NoMsg;
  return kvaDbOK;
}

//===========================================================================
// Get a signal given its name and its message.
KvaDbStatus WINAPI kvaDbGetSignalByName(KvaDbMessageHnd mh, char *buf, KvaDbSignalHnd *sh)
{
  if (!mh || !sh || !buf) return kvaDbErr_Param;
  CANdbMessage *m = (CANdbMessage*)mh;

  CANdbSignal *s = m->find_signal_by_name(buf);
  *sh = (KvaDbSignalHnd)s;
  if (!s) return kvaDbErr_NoMsg;
  return kvaDbOK;
}

//===========================================================================
// Tell, if a signal is a multiplexor, multiplexed or not multiplexed.
KvaDbStatus WINAPI kvaDbGetSignalMode(KvaDbSignalHnd sh, int *mux)
{
  if (!sh || !mux) return kvaDbErr_Param;
  CANdbSignal *s = (CANdbSignal*) sh;
  *mux = s->is_mode_signal()?KVADB_MUX_SIGNAL:s->get_mode();
  /* takes value KVADB_MUX_SIGNAL, if mux; KVADB_INDEPENDENT, if independent; otherwise returns mux mode value*/
  return kvaDbOK;
}

//===========================================================================
// Set signal's multiplexed value, updates mod_signal indicator
KvaDbStatus WINAPI kvaDbSetSignalMode(KvaDbSignalHnd sh, int mux)
{
  if (!sh || mux<-2) return kvaDbErr_Param;
  CANdbSignal *s = (CANdbSignal*) sh;
  if (mux<-1) {s->set_mode_signal(true);}
  else {s->set_mode_signal(false);}
  s->set_mode(mux);
  return kvaDbOK;
}

//===========================================================================
// Find a multiplexor signal for the given signal.
KvaDbStatus WINAPI kvaDbGetMsgMux(KvaDbMessageHnd mh, KvaDbSignalHnd *shh)
{
  if (!mh || !shh) return kvaDbErr_Param;
  CANdbMessage *msg = (CANdbMessage *) mh;

  (*shh) = (KvaDbSignalHnd) msg->get_mode_signal();
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalValueFloat(KvaDbSignalHnd sh, double *val,
                                            void *data, size_t len)
{
  int res = 0;
  if (!sh || !val || !data) return kvaDbErr_Param;
  if (len > 64) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;

  if (!sig->get_message()) return kvaDbErr_Param;

  switch (sig->get_type()) {
    case CANDB_SIGNED:
    {
      int value = 0;
      res = sig->get_value_int ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = sig->raw2phys((double)value);
      break;
    }
    case CANDB_UNSIGNED:
    {
      unsigned int value = 0;
      res = sig->get_value_uint ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = sig->raw2phys((double)value);
      break;
    }
    case CANDB_FLOAT:
    {
      double value = 0;
      res = sig->get_value_float ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = sig->raw2phys((double)value);
      break;
    }
    case CANDB_DOUBLE:
    {
      double value = 0;
      res = sig->get_value_double ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = sig->raw2phys((double)value);
      break;
    }
    default:
      return kvaDbErr_DatabaseInternal;
  }
  if (res != 0) return kvaDbErr_Param;
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalValueInteger(KvaDbSignalHnd sh, int *val,
                                              void *data, size_t len)
{
  int res = 0;
  if (!sh || !val || !data) return kvaDbErr_Param;
  if (len > 8) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;

  switch (sig->get_type()) {
    case CANDB_SIGNED:
    {
      int value = 0;
      res = sig->get_value_int ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = roundi(sig->raw2phys((double)value));
      break;
    }
    case CANDB_UNSIGNED:
    {
      unsigned int value = 0;
      res = sig->get_value_uint ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = roundu(sig->raw2phys((double)value));
      break;
    }
    case CANDB_FLOAT:
    {
      double value = 0;
      res = sig->get_value_float ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = roundi(sig->raw2phys((double)value));
      break;
    }
    case CANDB_DOUBLE:
    {
      double value = 0;
      res = sig->get_value_double ((unsigned char*)data, sig->get_message()->get_dlc(), value);
      *val = roundi(sig->raw2phys((double)value));
      break;
    }
    default:
      return kvaDbErr_DatabaseInternal;
  }
  if (res != 0) return kvaDbErr_Param;
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalValueEnum(KvaDbSignalHnd sh, char *buf,
                                            size_t buflen, void *data, size_t len)
{
  int val;
  if (!sh || !buf || !buflen || buflen == 0) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  if (!sig->has_symbolic_values()) return kvaDbErr_Fail;
  KvaDbStatus stat = kvaDbGetSignalValueInteger(sh, &val, data, len);
  if (stat != kvaDbOK) return stat;
  const char *name = sig->get_value_string(val);
  if (!name) return kvaDbErr_Fail;
  strncpy(buf, name, buflen);
  buf[buflen-1] = '\0';
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbStoreSignalValuePhys(KvaDbSignalHnd sh, unsigned char *can_data, int dlc, double value)
{
  if (!sh || !can_data || (dlc == 0)) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;

  double rawval = sig->phys2raw(value);
  if (sig->store_value(can_data, dlc, rawval)) {
    return kvaDbErr_Fail;
  }

  return kvaDbOK;
}
//===========================================================================
KvaDbStatus WINAPI kvaDbStoreSignalValueRaw(KvaDbSignalHnd sh, unsigned char *can_data, int dlc, int value)
{
  if (!sh || !can_data || (dlc == 0)) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;

  if (sig->store_value(can_data, dlc, value)) {
    return kvaDbErr_Fail;
  }

  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalValueLimits(KvaDbSignalHnd sh, double *minval, double *maxval)
{
  if (!sh || !minval || !maxval) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  *minval = sig->get_min_val();
  *maxval = sig->get_max_val();
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalValueScaling(KvaDbSignalHnd sh, double *factor, double *offset)
{
  if (!sh || !factor || !offset) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  *factor = sig->get_factor();
  *offset = sig->get_offset();
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalValueSize(KvaDbSignalHnd sh, int *startbit, int *length)
{
  if (!sh || !startbit || !length) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  *startbit = sig->get_start_bit();
  *length = sig->get_length();
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalName(KvaDbSignalHnd sh, char *buf, size_t buflen)
{
  if (!sh || !buf) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  strncpy(buf, sig->get_name(), buflen);
  buf[buflen-1] = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalQualifiedName(KvaDbSignalHnd sh, char *buf, size_t buflen)
{
  if (!sh || !buf) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->get_qualified_name(buf, (int) buflen);
  buf[buflen-1] = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalComment(KvaDbSignalHnd sh, char *buf, size_t buflen)
{
  if (!sh || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (sig && sig->get_comment()) strncpy(buf, sig->get_comment(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalUnit(KvaDbSignalHnd sh, char *buf, size_t buflen)
{
  if (!sh || !buf) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  //PRINTF(("sig=%p\n", sig));
  //PRINTF(("sig unit=%s\n", sig->get_unit()));
  if (sig && sig->get_unit()) {
    strncpy(buf, sig->get_unit(), buflen);
    buf[buflen-1] = 0;
  } else {
    if (buflen > 0) buf[0] = 0;
  }
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalEncoding(KvaDbSignalHnd sh, KvaDbSignalEncoding *e)
{
  if (!sh || !e) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (sig->is_motorola()) *e = kvaDb_Motorola;
  else *e = kvaDb_Intel;
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalRepresentationType(KvaDbSignalHnd sh, KvaDbSignalType *t)
{
  if (!sh || !t) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  switch (sig->get_type()) {
    case CANDB_SIGNED:      *t = kvaDb_Signed; break;
    case CANDB_UNSIGNED:    *t = kvaDb_Unsigned; break;
    case CANDB_FLOAT:       *t = kvaDb_Float; break;
    case CANDB_DOUBLE:      *t = kvaDb_Double; break;
    default:
      return kvaDbErr_DatabaseInternal;
  }
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbGetSignalPresentationType(KvaDbSignalHnd sh, KvaDbSignalType *t)
{
  if (!sh || !t) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  switch (sig->get_scaled_type()) {
    case CANDB_SIGNED:      *t = kvaDb_Signed; break;
    case CANDB_UNSIGNED:    *t = kvaDb_Unsigned; break;
    case CANDB_FLOAT:       *t = kvaDb_Float; break;
    case CANDB_DOUBLE:      *t = kvaDb_Double; break;
    default:
      return kvaDbErr_DatabaseInternal;
  }
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbAddSignal(KvaDbMessageHnd mh, KvaDbSignalHnd *sh)
{
  CANdbSignal *sig;
  CANdbMessage *m = (CANdbMessage*)mh;
  KvaDbStatus stat = kvaDbOK;

  if (!mh || !sh) return kvaDbErr_Param;
  sig = new CANdbSignal();
  if (m->insert_signal(sig) < 0) {
    stat = kvaDbErr_Param;
    delete sig;
    *sh = 0;
  } else {
    sig->set_name("");  // qqq or autogenerate?
    *sh = (KvaDbSignalHnd)sig;
  }
  return stat;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbDeleteSignal(KvaDbMessageHnd mh, KvaDbSignalHnd sh)
{
  KvaDbStatus stat = kvaDbOK;
  CANdbMessage *m = (CANdbMessage*)mh;
  CANdbSignal *s = (CANdbSignal*)sh;

  if (!sh || !mh) return kvaDbErr_Param;

  if (m->delete_signal(s) < 0) {
    stat = kvaDbErr_DatabaseInternal;
  }
  return stat;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalValueLimits(KvaDbSignalHnd sh, double minval, double maxval)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->set_min_val(minval);
  sig->set_max_val(maxval);
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalValueScaling(KvaDbSignalHnd sh, double factor, double offset)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->set_offset(offset);
  sig->set_factor(factor);
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalValueSize(KvaDbSignalHnd sh, int startbit, int length)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->set_start_bit(startbit);
  sig->set_length(length);
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalName(KvaDbSignalHnd sh, char *buf)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->set_name(buf);
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalComment(KvaDbSignalHnd sh, char *buf)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->set_comment(buf);
  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalUnit(KvaDbSignalHnd sh, char *buf)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  sig->set_unit(buf);
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalEncoding(KvaDbSignalHnd sh, KvaDbSignalEncoding e)
{
  if (!sh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (e == kvaDb_Motorola) sig->set_motorola(true);
  else sig->set_motorola(false);
  return kvaDbOK;
}


//===========================================================================
KvaDbStatus WINAPI kvaDbSetSignalRepresentationType(KvaDbSignalHnd sh, KvaDbSignalType t)
{
  if (!sh) return kvaDbErr_Param;

  CANdbSignal *sig = (CANdbSignal*)sh;
  switch (t) {
    case kvaDb_Signed:    sig->set_type(CANDB_SIGNED); break;
    case kvaDb_Unsigned:  sig->set_type(CANDB_UNSIGNED); break;
    case kvaDb_Float:     sig->set_type(CANDB_FLOAT); break;
    case kvaDb_Double:    sig->set_type(CANDB_DOUBLE); break;
    default:
      return kvaDbErr_Param;
  }

  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetFirstNode(KvaDbHnd dh, KvaDbNodeHnd *nh)
{
  if (!dh || !nh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster*)dh;

  CANdbNode *node = dc->get_first_node();
  if (!node) return kvaDbErr_NoNode;
  *nh = (KvaDbNodeHnd)node;

  return kvaDbOK;
}

//===========================================================================
KvaDbStatus WINAPI kvaDbGetNextNode(KvaDbHnd dh, KvaDbNodeHnd *nh)
{
  if (!dh || !nh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster*)dh;

  CANdbNode *node = (CANdbNode*)*nh;
  if (!node) return kvaDbErr_NoNode;

  CANdbNode *n = dc->get_next_node();
  *nh = (KvaDbNodeHnd)n;
  if (!*nh) return kvaDbErr_NoNode;

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNodeAttributeIntByName(KvaDbNodeHnd nh, const char *attrName, int *val)
{
  const CANdbAttribute *attr;

  if (!nh) return kvaDbErr_Param;
  CANdbNode *node = (CANdbNode*)nh;
  if (!node) return kvaDbErr_NoNode;

  CANdbAttributeList *aList = node->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  attr = aList->find_by_name(attrName);
  if (!attr) return kvaDbErr_NoAttrib;
  *val = attr->get_integer_value();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNodeName(KvaDbNodeHnd nh, char *buf, size_t buflen)
{
  if (!nh || !buf) return kvaDbErr_Param;
  CANdbNode *node = (CANdbNode*)nh;
  if (!node) return kvaDbErr_NoNode;

  strncpy(buf, node->get_name(), buflen);
  buf[buflen-1] = 0;

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNodeByName(KvaDbHnd dh, const char *node_name, KvaDbNodeHnd *nh)
{
  if (!dh || !nh) return kvaDbErr_Param;
  CANdbCluster *dc = (CANdbCluster *)dh;

  CANdbNode *n = dc->find_node_by_name(node_name);

  if (!n) return kvaDbErr_NoNode;
  *nh = (KvaDbNodeHnd)n;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddNode(KvaDbHnd dh, KvaDbNodeHnd *nh)
{
  CANdbCluster *dc = (CANdbCluster *)dh;

  if (!dh || !nh) return kvaDbErr_Param;

  CANdb *candb = dc->get_first_candb ();
  if (!candb) return kvaDbErr_NoDatabase;

  CANdbNode *node = new CANdbNode;

  if (candb->insert_node(node) < 0) {
    delete node;
    return kvaDbErr_Param;
  }

  *nh = (KvaDbNodeHnd)node;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbDeleteNode(KvaDbHnd dh, KvaDbNodeHnd nh)
{
  CANdbCluster *dc = (CANdbCluster *)dh;
  CANdbNode *node = (CANdbNode *)nh;

  if (!dh || !nh) return kvaDbErr_Param;

  CANdb *candb = dc->get_first_candb ();
  if (!candb) return kvaDbErr_NoDatabase;

  if (candb->remove_node(node) != 0) {
    return kvaDbErr_Param;
  }

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetNodeName(KvaDbNodeHnd nh, char *buf)
{
  CANdbNode *node = (CANdbNode *)nh;

  if (!node || !buf) return kvaDbErr_Param;

  node->set_name(buf);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetNodeComment(KvaDbNodeHnd nh, char *buf)
{
  CANdbNode *node = (CANdbNode *)nh;

  if (!node || !buf) return kvaDbErr_Param;

  node->set_comment(buf);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddReceiveNodeToSignal(KvaDbSignalHnd sh, KvaDbNodeHnd nh)
{
  CANdbSignal *sig = (CANdbSignal*)sh;
  CANdbNode *node = (CANdbNode *)nh;

  if (!sig || !node) return kvaDbErr_Param;

  sig->add_receive_node(node);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbRemoveReceiveNodeFromSignal(KvaDbSignalHnd sh, KvaDbNodeHnd nh)
{
  CANdbSignal *sig = (CANdbSignal*)sh;
  CANdbNode *node = (CANdbNode *)nh;

  if (!sig || !node) return kvaDbErr_Param;

  sig->remove_receive_node(node);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSignalContainsReceiveNode(KvaDbSignalHnd sh, KvaDbNodeHnd nh)
{
  CANdbSignal *sig = (CANdbSignal*)sh;
  CANdbNode *node = (CANdbNode *)nh;

  if (!sig || !node) return kvaDbErr_Param;

  if (!sig->contains_node(node)) return kvaDbErr_NoNode;

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddAttributeDefinition(KvaDbHnd dh, KvaDbAttributeDefHnd *adh)
{
  if (!dh) return kvaDbErr_Param;
  CANdbCluster * dc = (CANdbCluster *)dh;
  if (!dc) return kvaDbErr_NoDatabase;
  if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *db = dc->get_first_candb ();
  if (!db) return kvaDbErr_NoDatabase;

  CANdbAttributeDefinition *attr_def = new CANdbAttributeDefinition();
  *adh = (KvaDbAttributeDefHnd) attr_def;

  attr_def->set_name("");
  attr_def->set_type(CANDB_ATTR_TYPE_INVALID);
  attr_def->set_owner(CANDB_ATTR_OWNER_INVALID);

  db->insert_attribute_definition(attr_def);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionName(KvaDbAttributeDefHnd adh, const char *adName)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*) adh;

  attr_def->set_name(adName);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionType(KvaDbAttributeDefHnd adh, KvaDbAttributeType adType)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*) adh;

  attr_def->set_type((CANdbAttributeType) adType);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionOwner(KvaDbAttributeDefHnd adh, KvaDbAttributeOwner adOwner)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*) adh;

  attr_def->set_owner((CANdbAttributeOwner) adOwner);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddAttributeDefinitionEnum(KvaDbAttributeDefHnd adh, const char* eName, int eValue)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*) adh;

  return (KvaDbStatus)attr_def->add_enumeration(eName, eValue);
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionEnumDefault(KvaDbAttributeDefHnd adh, int def)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*) adh;

  attr_def->set_enumeration_default(def);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionInt(KvaDbAttributeDefHnd adh, int def, int min, int max)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;

  attr_def->set_integer_default(def);
  attr_def->set_integer_min(min);
  attr_def->set_integer_max(max);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionFloat(KvaDbAttributeDefHnd adh, float def, float min, float max)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;

  attr_def->set_float_default((double)def);
  attr_def->set_float_min((double)min);
  attr_def->set_float_max((double)max);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeDefinitionString(KvaDbAttributeDefHnd adh, const char *buf)
{
  if (!adh || !buf) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;

  attr_def->set_string_default(buf);

  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetFirstAttributeDefinition(KvaDbHnd dh, KvaDbAttributeDefHnd *adh)
{
  if (!dh) return kvaDbErr_Param;
  CANdbCluster * dc = (CANdbCluster *)dh;
  if (!dc) return kvaDbErr_NoDatabase;
  if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *db = dc->get_first_candb ();
  if (!db) return kvaDbErr_NoDatabase;

  CANdbAttributeDefinition *attr_def = db->get_first_attribute_definition();
  if (!attr_def) return kvaDbErr_NoAttrib;
  *adh = (KvaDbAttributeDefHnd)attr_def;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionByName(KvaDbHnd dh, const char *attrName, KvaDbAttributeDefHnd *adh)
{
  const CANdbAttributeDefinition *attr_def;

  if (!dh) return kvaDbErr_Param;
  CANdbCluster * dc = (CANdbCluster *)dh;
  if (!dc) return kvaDbErr_NoDatabase;
  if (dc->get_candb_count () > 1) return kvaDbErr_OnlyOneAllowed;
  CANdb *db = dc->get_first_candb ();
  if (!db) return kvaDbErr_NoDatabase;

  attr_def = db->find_attribute_definition_by_name(attrName);
  if (!attr_def) return kvaDbErr_NoAttrib;
  *adh = (KvaDbAttributeDefHnd)attr_def;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNextAttributeDefinition(KvaDbAttributeDefHnd adh, KvaDbAttributeDefHnd *nadh)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  CANdbAttributeDefinition *n_attr_def = attr_def->get_next();
  if (!n_attr_def) return kvaDbErr_NoAttrib;
  *nadh = (KvaDbAttributeDefHnd)n_attr_def;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionType(KvaDbAttributeDefHnd adh, KvaDbAttributeType *at)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  *at = (KvaDbAttributeType)attr_def->get_type();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionOwner(KvaDbAttributeDefHnd adh, KvaDbAttributeOwner *ao)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  *ao = (KvaDbAttributeOwner)attr_def->get_owner();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionName(KvaDbAttributeDefHnd adh, char *buf, size_t buflen)
{
  if (!adh || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;

  if (attr_def && attr_def->get_name()) strncpy(buf, attr_def->get_name(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionInt(KvaDbAttributeDefHnd adh, int *def, int *min, int *max)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  *def = attr_def->get_integer_default();
  *min = attr_def->get_integer_min();
  *max = attr_def->get_integer_max();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionFloat(KvaDbAttributeDefHnd adh, float *def, float *min, float *max)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  *def = (float)attr_def->get_float_default();
  *min = (float)attr_def->get_float_min();
  *max = (float)attr_def->get_float_max();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionString(KvaDbAttributeDefHnd adh, char *buf, size_t buflen)
{
  if (!adh || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;

  if (attr_def && attr_def->get_string_default()) strncpy(buf, attr_def->get_string_default(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeDefinitionEnumeration(KvaDbAttributeDefHnd adh, int *def)
{
  if (!adh) return kvaDbErr_Param;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  *def = (int)attr_def->get_enumeration_default();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddMsgAttribute(KvaDbMessageHnd mh, KvaDbAttributeDefHnd adh, KvaDbAttributeHnd *ah)
{
  if (!mh || !adh) return kvaDbErr_Param;
  CANdbMessage *msg = (CANdbMessage*)mh;
  if (!msg) return kvaDbErr_NoMsg;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  CANdbAttributeOwner ao = attr_def->get_owner();
  if (ao != CANDB_ATTR_OWNER_MESSAGE) return kvaDbErr_Fail;
  CANdbAttributeList *aList = msg->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  CANdbAttribute *a_tmp = aList->find_by_definition(attr_def);
  if (a_tmp) return kvaDbErr_Fail;
  CANdbAttribute *attr = new CANdbAttribute(attr_def);
  aList->insert(attr);
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbDeleteMsgAttribute(KvaDbMessageHnd mh, KvaDbAttributeHnd ah)
{
  if (!mh || !ah) return kvaDbErr_Param;
  CANdbMessage *msg = (CANdbMessage*)mh;
  if (!msg) return kvaDbErr_NoMsg;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  CANdbAttributeList *aList = msg->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  aList->remove(attr);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddSignalAttribute(KvaDbSignalHnd sh, KvaDbAttributeDefHnd adh, KvaDbAttributeHnd *ah)
{
  if (!sh || !adh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  CANdbAttributeOwner ao = attr_def->get_owner();
  if (ao != CANDB_ATTR_OWNER_SIGNAL) return kvaDbErr_Fail;
  CANdbAttributeList *aList = sig->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  CANdbAttribute *a_tmp = aList->find_by_definition(attr_def);
  if (a_tmp) return kvaDbErr_Fail;
  CANdbAttribute *attr = new CANdbAttribute(attr_def);
  aList->insert(attr);
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetFirstEnumValue(KvaDbSignalHnd sh, KvaDbEnumValueHnd *eh)
{
  if (!sh || !eh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  CANdbEnumValue *e = sig->get_first_value();
  *eh = (KvaDbEnumValueHnd)e;
  if (!e) return kvaDbErr_NoAttrib;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNextEnumValue(KvaDbSignalHnd sh, KvaDbEnumValueHnd *eh)
{
  if (!sh || !eh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  CANdbEnumValue *e = sig->get_next_value();
  *eh = (KvaDbEnumValueHnd)e;
  if (!e) return kvaDbErr_NoAttrib;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetEnumValue(KvaDbEnumValueHnd eh, int *val, char *buf, size_t buflen)
{
  if (!eh || !val || !buf || !buflen) return kvaDbErr_Param;
  CANdbEnumValue *e = (CANdbEnumValue*)eh;
  if (!e) return kvaDbErr_NoAttrib;
  *val = e->get_value();
  strncpy(buf, e->get_name(), buflen);
  buf[buflen-1] = 0;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddEnumValue(KvaDbSignalHnd sh, int val, const char* name)
{
  if (!sh || !name) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  if (sig->add_value(val, name)) return kvaDbErr_Fail;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbDeleteEnumValue(KvaDbSignalHnd sh, KvaDbEnumValueHnd eh)
{
  if (!sh || !eh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  CANdbEnumValue *e = (CANdbEnumValue*)eh;
  if (!e) return kvaDbErr_NoAttrib;
  if (sig->remove_value(e)) return kvaDbErr_Fail;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbDeleteSignalAttribute(KvaDbSignalHnd sh, KvaDbAttributeHnd ah)
{
  if (!sh || !ah) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  CANdbAttributeList *aList = sig->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  aList->remove(attr);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbAddNodeAttribute(KvaDbNodeHnd nh, KvaDbAttributeDefHnd adh, KvaDbAttributeHnd *ah)
{
  if (!nh || !adh) return kvaDbErr_Param;
  CANdbNode *n = (CANdbNode*)nh;
  if (!n) return kvaDbErr_NoNode;
  CANdbAttributeDefinition *attr_def = (CANdbAttributeDefinition*)adh;
  if (!attr_def) return kvaDbErr_NoAttrib;
  CANdbAttributeOwner ao = attr_def->get_owner();
  if (ao != CANDB_ATTR_OWNER_NODE) return kvaDbErr_Fail;
  CANdbAttributeList *aList = n->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  CANdbAttribute *a_tmp = aList->find_by_definition(attr_def);
  if (a_tmp) return kvaDbErr_Fail;
  CANdbAttribute *attr = new CANdbAttribute(attr_def);
  aList->insert(attr);
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbDeleteNodeAttribute(KvaDbNodeHnd nh, KvaDbAttributeHnd ah)
{
  if (!nh || !ah) return kvaDbErr_Param;
  CANdbNode *n = (CANdbNode*)nh;
  if (!n) return kvaDbErr_NoNode;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  CANdbAttributeList *aList = n->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  aList->remove(attr);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetFirstMsgAttribute(KvaDbMessageHnd mh, KvaDbAttributeHnd *ah)
{
  if (!mh) return kvaDbErr_Param;
  CANdbMessage *msg = (CANdbMessage*)mh;
  if (!msg) return kvaDbErr_NoMsg;

  CANdbAttributeList *aList = msg->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  CANdbAttribute *attr = aList->get_first_attribute();
  if (!attr) return kvaDbErr_NoAttrib;
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetMsgAttributeByName(KvaDbMessageHnd mh, const char *attrName, KvaDbAttributeHnd *ah)
{
  const CANdbAttribute *attr;

  if (!mh) return kvaDbErr_Param;
  CANdbMessage *msg = (CANdbMessage*)mh;
  if (!msg) return kvaDbErr_NoMsg;

  CANdbAttributeList *aList = msg->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  attr = aList->find_by_name(attrName);
  if (!attr) return kvaDbErr_NoAttrib;
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetFirstSignalAttribute(KvaDbSignalHnd sh, KvaDbAttributeHnd *ah)
{
  if (!sh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;

  CANdbAttributeList *aList = sig->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  CANdbAttribute *attr = aList->get_first_attribute();
  if (!attr) return kvaDbErr_NoAttrib;
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetSignalAttributeByName(KvaDbSignalHnd sh, const char *attrName, KvaDbAttributeHnd *ah)
{
  const CANdbAttribute *attr;

  if (!sh) return kvaDbErr_Param;
  CANdbSignal *sig = (CANdbSignal*)sh;
  if (!sig) return kvaDbErr_NoSignal;

  CANdbAttributeList *aList = sig->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  attr = aList->find_by_name(attrName);
  if (!attr) return kvaDbErr_NoAttrib;
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetFirstNodeAttribute(KvaDbNodeHnd nh, KvaDbAttributeHnd *ah)
{
  if (!nh) return kvaDbErr_Param;
  CANdbNode *n = (CANdbNode*)nh;
  if (!n) return kvaDbErr_NoNode;

  CANdbAttributeList *aList = n->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  CANdbAttribute *attr = aList->get_first_attribute();
  if (!attr) return kvaDbErr_NoAttrib;
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNodeAttributeByName(KvaDbNodeHnd nh, const char *attrName, KvaDbAttributeHnd *ah)
{
  const CANdbAttribute *attr;

  if (!nh) return kvaDbErr_Param;
  CANdbNode *n = (CANdbNode*)nh;
  if (!n) return kvaDbErr_NoNode;

  CANdbAttributeList *aList = n->get_attributes();
  if (!aList) return kvaDbErr_NoAttrib;
  attr = aList->find_by_name(attrName);
  if (!attr) return kvaDbErr_NoAttrib;
  *ah = (KvaDbAttributeHnd)attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetNextAttribute(KvaDbAttributeHnd ah, KvaDbAttributeHnd *nah)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  CANdbAttribute *n_attr = attr->get_next();
  if (!n_attr) return kvaDbErr_NoAttrib;
  *nah = (KvaDbAttributeHnd)n_attr;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeType(KvaDbAttributeHnd ah, KvaDbAttributeType *at)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  *at = (KvaDbAttributeType)attr->get_type();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeName(KvaDbAttributeHnd ah, char *buf, size_t buflen)
{
  if (!ah || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  CANdbAttributeDefinition *attr_def = attr->get_definition();
  if (!attr_def) return kvaDbErr_NoAttrib;

  if (attr_def && attr_def->get_name()) strncpy(buf, attr_def->get_name(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeValueInt(KvaDbAttributeHnd ah, int val)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  attr->set_integer_value(val);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeValueInt(KvaDbAttributeHnd ah, int *val)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  *val = attr->get_integer_value();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeValueFloat(KvaDbAttributeHnd ah, float val)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  attr->set_float_value(val);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeValueFloat(KvaDbAttributeHnd ah, float *val)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  *val = (float)attr->get_float_value();
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeValueString(KvaDbAttributeHnd ah, const char *buf, int /* buflen */)
{
  if (!ah || !buf) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  attr->set_string_value(buf);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeValueString(KvaDbAttributeHnd ah, char *buf, size_t buflen)
{
  if (!ah || !buf || (buflen == 0)) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;

  if (attr && attr->get_string_value()) strncpy(buf, attr->get_string_value(), buflen);
  else strcpy (buf, "");
  buf[buflen-1] = 0;
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbSetAttributeValueEnumeration(KvaDbAttributeHnd ah, int val)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  attr->set_enumeration_value(val);
  return kvaDbOK;
}

//===========================================================================

KvaDbStatus WINAPI kvaDbGetAttributeValueEnumeration(KvaDbAttributeHnd ah, int *val)
{
  if (!ah) return kvaDbErr_Param;
  CANdbAttribute *attr = (CANdbAttribute*)ah;
  if (!attr) return kvaDbErr_NoAttrib;
  *val = (int)attr->get_enumeration_value();
  return kvaDbOK;
}

//===========================================================================
